package yaes.ui.propertysheet;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.util.Collections;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JComponent;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.border.Border;
import javax.swing.border.EmptyBorder;

/**
 * Common base class for dialogs. Provides simple layout (vertical order of
 * elements, paired with label). Also, offers set of basic building blocks for
 * more complex dialogs.
 */
public class DialogHelper {
	public final static int VGAP = 10;
	public final static int HGAP = 30;
	// external spacing
	public final static int ESPACE = 5;

	/**
	 * Adding a check box to specified panel
	 * 
	 * @param panel
	 *            panel onto which add label + check box
	 * @param label
	 *            text of label
	 * @return check box object
	 */
	public static JCheckBox addCheckBox(JPanel panel, String label) {
		if (label != null) {
			final JLabel lb = new JLabel(label, SwingConstants.LEFT);
			panel.add(lb);
		}
		final JCheckBox cb = new JCheckBox();
		panel.add(cb);
		return cb;
	}

	/**
	 * Adding a combobox to specified panel
	 * 
	 * @param panel
	 *            panel onto which add label + combo box
	 * @param label
	 *            text of label
	 * @return combo box object
	 */
	public static JComboBox addComboBox(JPanel panel, String label) {
		if (label != null) {
			final JLabel lb = new JLabel(label, SwingConstants.LEFT);
			panel.add(lb);
		}
		final JComboBox cb = new JComboBox();
		cb.setEditable(true);
		panel.add(cb);
		return cb;
	}

	/**
	 * Utility functions for adding stuff
	 */
	/**
	 * Adding a label
	 * 
	 * @param panel
	 *            panel onto which add label
	 * @param label
	 *            text of label
	 * @return label object
	 */
	public static JLabel addLabel(JPanel panel, String label) {
		final JLabel lb = new JLabel(label, SwingConstants.LEFT);
		panel.add(lb);
		return lb;
	}

	/**
	 * Adding a listbox to specified panel
	 * 
	 * @param panel
	 *            panel onto which add label + list
	 * @param label
	 *            text of label
	 * @return list box object
	 */
	public static JList addList(JPanel panel, String label) {
		if (label != null) {
			final JLabel lb = new JLabel(label, SwingConstants.LEFT);
			panel.add(lb);
		}
		final JList lb = new JList();
		panel.add(lb);
		return lb;
	}

	/**
	 * Adding a text area to specified panel
	 * 
	 * @param panel
	 *            panel onto which add label + text area
	 * @param label
	 *            text of label
	 * @param width
	 *            width of text area
	 * @param height
	 *            height of text area
	 * @return text area object
	 */
	public static JTextArea addTextArea(JPanel panel, String label, int width,
			int height) {
		if (label != null) {
			final JLabel lb = new JLabel(label, SwingConstants.LEFT);
			panel.add(lb);
		}
		final JTextArea ta = new JTextArea(width, height);
		final JScrollPane scroll = new JScrollPane(ta);
		// scroll.getViewport().setView(ta);
		panel.add(scroll);
		return ta;
	}

	/**
	 * Adds to specified panel: text field, preceded by label Size of text field
	 * is system default
	 * 
	 * @param panel
	 *            panel onto which add label + field
	 * @param label
	 *            text of label
	 * @return text field object
	 */
	public static JTextField addTextField(JPanel panel, String label) {
		if (label != null) {
			final JLabel lb = new JLabel(label, SwingConstants.LEFT);
			panel.add(lb);
		}
		final JTextField tf = new JTextField();
		panel.add(tf);
		return tf;
	}

	/**
	 * Adds to specified panel: text field, preceded by label Size of text field
	 * ii specified by width
	 * 
	 * @param panel
	 *            panel onto which add label + field
	 * @param label
	 *            text of label
	 * @param width
	 *            width of text field
	 * @return text field object
	 */
	public static JTextField addTextField(JPanel panel, String label, int width) {
		if (label != null) {
			final JLabel lb = new JLabel(label, SwingConstants.LEFT);
			panel.add(lb);
		}
		final JTextField tf = new JTextField(width);
		panel.add(tf);
		return tf;
	}

	public static Border createBalancedEmptyBorder(boolean insert_space_top,
			boolean insert_space_left, boolean insert_space_bottom,
			boolean insert_space_right) {
		final int top = insert_space_top ? DialogHelper.ESPACE : 0;
		final int left = insert_space_left ? DialogHelper.ESPACE : 0;
		final int bottom = insert_space_bottom ? DialogHelper.ESPACE : 0;
		final int right = insert_space_right ? DialogHelper.ESPACE : 0;
		return BorderFactory.createEmptyBorder(top, left, bottom, right);
	}

	// Border combining titled border and empty border
	public static Border createBalancedInnerBorder(String title,
			boolean insert_space_top, boolean insert_space_left,
			boolean insert_space_bottom, boolean insert_space_right) {
		if (title == null) {
			title = "";
		}
		return BorderFactory.createCompoundBorder(BorderFactory
				.createTitledBorder(BorderFactory.createEtchedBorder(), title),
				DialogHelper.createBalancedEmptyBorder(insert_space_top,
						insert_space_left, insert_space_bottom,
						insert_space_right));
	}

	public static void createDualPanel(JPanel basePanel, JPanel labels,
			JPanel fields) {
		basePanel.setBorder(new EmptyBorder(20, 20, 20, 20));
		basePanel.setLayout(new BorderLayout());
		final JPanel pushUp = new JPanel();
		pushUp.setLayout(new BorderLayout());
		basePanel.add(pushUp, BorderLayout.NORTH);
		final JPanel doCenter = new JPanel();
		doCenter.setLayout(new BorderLayout());
		pushUp.add(doCenter, BorderLayout.CENTER);
		final GridLayout labelgl = new GridLayout(0, 1);
		labelgl.setVgap(5);
		labels.setLayout(labelgl);
		doCenter.add(labels, BorderLayout.CENTER);
		final GridLayout fieldgl = new GridLayout(0, 1);
		fieldgl.setVgap(5);
		fields.setLayout(fieldgl);
		doCenter.add(fields, BorderLayout.EAST);
	}

	/**
	 * Created panel consists of labels and fields panel with hgap space between
	 * them labels and fields panels should provide their own layouts
	 */
	public static void createFormPanel(JPanel basePanel, JPanel labels,
			JPanel fields, int hGap) {
		final BorderLayout bl = new BorderLayout();
		basePanel.setLayout(bl);
		final JPanel outerLabels = new JPanel();
		outerLabels.setLayout(new BorderLayout());
		outerLabels.add(labels);
		outerLabels.add(Box.createHorizontalStrut(hGap), BorderLayout.EAST);
		basePanel.add(outerLabels, BorderLayout.WEST);
		basePanel.add(fields, BorderLayout.CENTER);
	}

	public static Border createInnerBorder(String title) {
		if (title == null) {
			title = "";
		}
		return BorderFactory.createCompoundBorder(BorderFactory
				.createTitledBorder(BorderFactory.createEtchedBorder(), title),
				DialogHelper.createBalancedEmptyBorder(true, true, true, true));
	}

	public static Border createOuterBorder(Color color) {
		if (color == null) {
			color = Color.black;
		}
		return BorderFactory.createCompoundBorder(BorderFactory
				.createEtchedBorder(color, Color.GRAY), BorderFactory
				.createEmptyBorder(DialogHelper.VGAP, DialogHelper.VGAP,
						DialogHelper.VGAP, DialogHelper.VGAP));
	}

	public static Border createStandardEmptyBorder() {
		return DialogHelper.createBalancedEmptyBorder(true, true, true, true);
	}

	/**
	 * Description of the Method
	 */
	public static int getWidestComponentWidth(Vector components) {
		final Vector<Integer> widths = new Vector<Integer>();
		int largestCompWidth = 0;
		if (components == null) {
			return largestCompWidth;
		}
		final Iterator<?> iter = components.iterator();
		while (iter.hasNext()) {
			final Object obj = iter.next();
			if (obj instanceof JComponent) {
				widths.add(new Integer(
						((JComponent) obj).getPreferredSize().width));
			}
		}
		try {
			final Integer max = Collections.max(widths);
			largestCompWidth = max.intValue();
		} catch (final Exception exc) {
		}
		return largestCompWidth;
	}

	public static void setDialogLocation(JDialog dialog) {
		if (dialog == null) {
			return;
		}
		final Window win = dialog.getOwner();
		if (win == null) {
			final Point pt = new Point(100, 100);
			dialog.setLocation(pt);
			return;
		}
		final Dimension dim1 = win.getSize();
		final Point pt = win.getLocation();
		final Dimension dim2 = dialog.getSize();
		pt.translate(dim1.width / 2 - dim2.width / 2, dim1.height / 2
				- dim2.height / 2);
		if (pt.x < 0) {
			pt.x = 0;
		}
		if (pt.y < 0) {
			pt.y = 0;
		}
		dialog.setLocation(pt);
	}
}

/**
 * Encapsulates use of file select dialog
 */
class FileSelectorHandler implements ActionListener {
	JTextField textField;
	JDialog parent;

	/**
	 * @param parent
	 * @param textField
	 */
	FileSelectorHandler(JDialog parent, JTextField textField) {
		this.textField = textField;
		this.parent = parent;
	}

	/**
	 * @param e
	 */
	public void actionPerformed(ActionEvent e) {
		final JFileChooser fc = new JFileChooser();
		// XMLFileFilter filter = new XMLFileFilter();
		// filter.setDescription("JPG & GIF Images");
		// fc.addChoosableFileFilter(filter);
		// fc.addChoosableFileFilter(fc.getAcceptAllFileFilter());
		// fc.setFileFilter(filter);
		final int returnVal = fc.showOpenDialog(parent);
		if (returnVal == JFileChooser.APPROVE_OPTION) {
			final File file = fc.getSelectedFile();
			textField.setText(file.getAbsolutePath());
		}
		// open
	}
}
